@@ -1,5 +1,5 @@
 /******************************************************************************
- * $Id: vrtdataset.h 56578e74c2851abf597312e07504cfe639cceedd 2018-04-24 20:38:09 +0200 Even Rouault $
+ * $Id: vrtdataset.h c3bfdaa3118385908c35a20b0d4e96de3322c77f 2019-11-15 13:58:59 +0100 Even Rouault $
  *
  * Project:  Virtual GDAL Datasets
  * Purpose:  Declaration of virtual gdal dataset classes.
@@ -38,6 +38,7 @@
 #include "gdal_priv.h"
 #include "gdal_rat.h"
 #include "gdal_vrt.h"
+#include "gdal_rat.h"
 
 #include <map>
 #include <memory>
@@ -60,13 +61,24 @@ void* VRTDeserializeWarpedOverviewTransformer( CPLXMLNode *psTree );
 /************************************************************************/
 class VRTOverviewInfo
 {
+    CPL_DISALLOW_COPY_ASSIGN(VRTOverviewInfo)
+
 public:
-    CPLString       osFilename;
-    int             nBand;
-    GDALRasterBand *poBand;
-    int             bTriedToOpen;
+    CPLString       osFilename{};
+    int             nBand = 0;
+    GDALRasterBand *poBand = nullptr;
+    int             bTriedToOpen = FALSE;
+
+    VRTOverviewInfo() = default;
+    VRTOverviewInfo(VRTOverviewInfo&& oOther) noexcept:
+        osFilename(std::move(oOther.osFilename)),
+        nBand(oOther.nBand),
+        poBand(oOther.poBand),
+        bTriedToOpen(oOther.bTriedToOpen)
+    {
+        oOther.poBand = nullptr;
+    }
 
-    VRTOverviewInfo() : nBand(0), poBand(nullptr), bTriedToOpen(FALSE) {}
     ~VRTOverviewInfo() {
         if( poBand == nullptr )
             /* do nothing */;
@@ -110,7 +122,8 @@ public:
                                   GDALProgressFunc pfnProgress,
                                   void *pProgressData ) = 0;
 
-    virtual CPLErr  XMLInit( CPLXMLNode *psTree, const char *, void* ) = 0;
+    virtual CPLErr  XMLInit( CPLXMLNode *psTree, const char *, void*,
+                             std::map<CPLString, GDALDataset*>& ) = 0;
     virtual CPLXMLNode *SerializeToXML( const char *pszVRTPath ) = 0;
 
     virtual void   GetFileList(char*** ppapszFileList, int *pnSize,
@@ -120,10 +133,13 @@ public:
     virtual CPLErr FlushCache() { return CE_None; }
 };
 
-typedef VRTSource *(*VRTSourceParser)(CPLXMLNode *, const char *, void* pUniqueHandle);
+typedef VRTSource *(*VRTSourceParser)(CPLXMLNode *, const char *, void* pUniqueHandle,
+                                      std::map<CPLString, GDALDataset*>& oMapSharedSources);
 
-VRTSource *VRTParseCoreSources( CPLXMLNode *psTree, const char *, void* pUniqueHandle );
-VRTSource *VRTParseFilterSources( CPLXMLNode *psTree, const char *, void* pUniqueHandle );
+VRTSource *VRTParseCoreSources( CPLXMLNode *psTree, const char *, void* pUniqueHandle,
+                                std::map<CPLString, GDALDataset*>& oMapSharedSources);
+VRTSource *VRTParseFilterSources( CPLXMLNode *psTree, const char *, void* pUniqueHandle,
+                                  std::map<CPLString, GDALDataset*>& oMapSharedSources );
 
 /************************************************************************/
 /*                              VRTDataset                              */
@@ -131,18 +147,30 @@ VRTSource *VRTParseFilterSources( CPLXMLNode *psTree, const char *, void* pUniqu
 
 class VRTRasterBand;
 
+template<class T> struct VRTFlushCacheStruct
+{
+    static void FlushCache(T& obj);
+};
+
+class VRTWarpedDataset;
+class VRTPansharpenedDataset;
+
 class CPL_DLL VRTDataset : public GDALDataset
 {
     friend class VRTRasterBand;
+    friend struct VRTFlushCacheStruct<VRTDataset>;
+    friend struct VRTFlushCacheStruct<VRTWarpedDataset>;
+    friend struct VRTFlushCacheStruct<VRTPansharpenedDataset>;
+    friend class VRTSourcedRasterBand;
 
-    char           *m_pszProjection;
+    OGRSpatialReference* m_poSRS = nullptr;
 
     int            m_bGeoTransformSet;
     double         m_adfGeoTransform[6];
 
     int            m_nGCPCount;
     GDAL_GCP      *m_pasGCPList;
-    char          *m_pszGCPProjection;
+    OGRSpatialReference *m_poGCP_SRS = nullptr;
 
     int            m_bNeedsFlush;
     int            m_bWritable;
@@ -153,13 +181,19 @@ class CPL_DLL VRTDataset : public GDALDataset
 
     int            m_bCompatibleForDatasetIO;
     int            CheckCompatibleForDatasetIO();
+    void           ExpandProxyBands();
+
     std::vector<GDALDataset*> m_apoOverviews;
     std::vector<GDALDataset*> m_apoOverviewsBak;
     char         **m_papszXMLVRTMetadata;
 
+    std::map<CPLString, GDALDataset*> m_oMapSharedSources;
+
     VRTRasterBand*      InitBand(const char* pszSubclass, int nBand,
                                  bool bAllowPansharpened);
 
+    CPL_DISALLOW_COPY_ASSIGN(VRTDataset)
+
   protected:
     virtual int         CloseDependentDatasets() override;
 
@@ -175,8 +209,9 @@ class CPL_DLL VRTDataset : public GDALDataset
     virtual CPLErr          CreateMaskBand( int nFlags ) override;
     void SetMaskBand(VRTRasterBand* poMaskBand);
 
-    virtual const char *GetProjectionRef() override;
-    virtual CPLErr SetProjection( const char * ) override;
+    const OGRSpatialReference* GetSpatialRef() const override { return m_poSRS; }
+    CPLErr SetSpatialRef(const OGRSpatialReference* poSRS) override;
+
     virtual CPLErr GetGeoTransform( double * ) override;
     virtual CPLErr SetGeoTransform( double * ) override;
 
@@ -188,10 +223,11 @@ class CPL_DLL VRTDataset : public GDALDataset
     virtual char** GetMetadata( const char *pszDomain = "" ) override;
 
     virtual int    GetGCPCount() override;
-    virtual const char *GetGCPProjection() override;
+    const OGRSpatialReference* GetGCPSpatialRef() const override { return m_poGCP_SRS; }
     virtual const GDAL_GCP *GetGCPs() override;
-    virtual CPLErr SetGCPs( int nGCPCount, const GDAL_GCP *pasGCPList,
-                            const char *pszGCPProjection ) override;
+    using GDALDataset::SetGCPs;
+    CPLErr SetGCPs( int nGCPCount, const GDAL_GCP *pasGCPList,
+                    const OGRSpatialReference* poSRS ) override;
 
     virtual CPLErr AddBand( GDALDataType eType,
                             char **papszOptions=nullptr ) override;
@@ -266,6 +302,8 @@ class CPL_DLL VRTWarpedDataset : public VRTDataset
 
     friend class VRTWarpedRasterBand;
 
+    CPL_DISALLOW_COPY_ASSIGN(VRTWarpedDataset)
+
   protected:
     virtual int         CloseDependentDatasets() override;
 
@@ -273,6 +311,8 @@ public:
                       VRTWarpedDataset( int nXSize, int nYSize );
     virtual ~VRTWarpedDataset();
 
+    virtual void  FlushCache() override;
+
     CPLErr            Initialize( /* GDALWarpOptions */ void * );
 
     virtual CPLErr IBuildOverviews( const char *, int, int *,
@@ -340,6 +380,8 @@ class VRTPansharpenedDataset : public VRTDataset
 
     std::vector<GDALDataset*> m_apoDatasetsToClose;
 
+    CPL_DISALLOW_COPY_ASSIGN(VRTPansharpenedDataset)
+
   protected:
     virtual int         CloseDependentDatasets() override;
 
@@ -347,6 +389,8 @@ public:
                       VRTPansharpenedDataset( int nXSize, int nYSize );
     virtual ~VRTPansharpenedDataset();
 
+    virtual void  FlushCache() override;
+
     virtual CPLErr    XMLInit( CPLXMLNode *, const char * ) override;
     virtual CPLXMLNode *   SerializeToXML( const char *pszVRTPath ) override;
 
@@ -411,12 +455,15 @@ class CPL_DLL VRTRasterBand : public GDALRasterBand
 
     std::unique_ptr<GDALRasterAttributeTable> m_poRAT;
 
+    CPL_DISALLOW_COPY_ASSIGN(VRTRasterBand)
+
   public:
 
                     VRTRasterBand();
     virtual        ~VRTRasterBand();
 
-    virtual CPLErr         XMLInit( CPLXMLNode *, const char *, void* );
+    virtual CPLErr         XMLInit( CPLXMLNode *, const char *, void*,
+                                    std::map<CPLString, GDALDataset*>& );
     virtual CPLXMLNode *   SerializeToXML( const char *pszVRTPath );
 
     virtual CPLErr SetNoDataValue( double ) override;
@@ -503,6 +550,8 @@ class CPL_DLL VRTSourcedRasterBand : public VRTRasterBand
     bool           CanUseSourcesMinMaxImplementations();
     void           CheckSource( VRTSimpleSource *poSS );
 
+    CPL_DISALLOW_COPY_ASSIGN(VRTSourcedRasterBand)
+
   public:
     int            nSources;
     VRTSource    **papoSources;
@@ -536,7 +585,8 @@ class CPL_DLL VRTSourcedRasterBand : public VRTRasterBand
                                          const char * pszValue,
                                          const char * pszDomain = "" ) override;
 
-    virtual CPLErr         XMLInit( CPLXMLNode *, const char *, void* ) override;
+    virtual CPLErr         XMLInit( CPLXMLNode *, const char *, void*,
+                                    std::map<CPLString, GDALDataset*>& ) override;
     virtual CPLXMLNode *   SerializeToXML( const char *pszVRTPath ) override;
 
     virtual double GetMinimum( int *pbSuccess = nullptr ) override;
@@ -668,6 +718,8 @@ class CPL_DLL VRTDerivedRasterBand : public VRTSourcedRasterBand
     VRTDerivedRasterBandPrivateData* m_poPrivate;
     bool InitializePython();
 
+    CPL_DISALLOW_COPY_ASSIGN(VRTDerivedRasterBand)
+
  public:
     char *pszFuncName;
     GDALDataType eSourceTransferType;
@@ -695,7 +747,8 @@ class CPL_DLL VRTDerivedRasterBand : public VRTSourcedRasterBand
     void SetSourceTransferType( GDALDataType eDataType );
     void SetPixelFunctionLanguage( const char* pszLanguage );
 
-    virtual CPLErr         XMLInit( CPLXMLNode *, const char *, void* ) override;
+    virtual CPLErr         XMLInit( CPLXMLNode *, const char *, void*,
+                                    std::map<CPLString, GDALDataset*>& ) override;
     virtual CPLXMLNode *   SerializeToXML( const char *pszVRTPath ) override;
 
     virtual double GetMinimum( int *pbSuccess = nullptr ) override;
@@ -728,12 +781,15 @@ class CPL_DLL VRTRawRasterBand : public VRTRasterBand
     char           *m_pszSourceFilename;
     int            m_bRelativeToVRT;
 
+    CPL_DISALLOW_COPY_ASSIGN(VRTRawRasterBand)
+
   public:
                    VRTRawRasterBand( GDALDataset *poDS, int nBand,
                                      GDALDataType eType = GDT_Unknown );
     virtual        ~VRTRawRasterBand();
 
-    virtual CPLErr         XMLInit( CPLXMLNode *, const char *, void* ) override;
+    virtual CPLErr         XMLInit( CPLXMLNode *, const char *, void*,
+                                    std::map<CPLString, GDALDataset*>& ) override;
     virtual CPLXMLNode *   SerializeToXML( const char *pszVRTPath ) override;
 
     virtual CPLErr IRasterIO( GDALRWFlag, int, int, int, int,
@@ -763,6 +819,8 @@ class CPL_DLL VRTRawRasterBand : public VRTRasterBand
 
 class VRTDriver : public GDALDriver
 {
+    CPL_DISALLOW_COPY_ASSIGN(VRTDriver)
+
   public:
                  VRTDriver();
     virtual ~VRTDriver();
@@ -775,7 +833,8 @@ class VRTDriver : public GDALDriver
                                      const char * pszDomain = "" ) override;
 
     VRTSource   *ParseSource( CPLXMLNode *psSrc, const char *pszVRTPath,
-                              void* pUniqueHandle );
+                              void* pUniqueHandle,
+                              std::map<CPLString, GDALDataset*>& oMapSharedSources );
     void         AddSourceParser( const char *pszElementName,
                                   VRTSourceParser pfnParser );
 };
@@ -786,6 +845,8 @@ class VRTDriver : public GDALDriver
 
 class CPL_DLL VRTSimpleSource : public VRTSource
 {
+    CPL_DISALLOW_COPY_ASSIGN(VRTSimpleSource)
+
 protected:
     friend class VRTSourcedRasterBand;
 
@@ -823,7 +884,8 @@ public:
                              double dfXDstRatio, double dfYDstRatio );
     virtual ~VRTSimpleSource();
 
-    virtual CPLErr  XMLInit( CPLXMLNode *psTree, const char *, void* ) override;
+    virtual CPLErr  XMLInit( CPLXMLNode *psTree, const char *, void*,
+                             std::map<CPLString, GDALDataset*>& ) override;
     virtual CPLXMLNode *SerializeToXML( const char *pszVRTPath ) override;
 
     void           SetSrcBand( GDALRasterBand * );
@@ -877,6 +939,7 @@ public:
     virtual CPLErr FlushCache() override;
 
     GDALRasterBand* GetBand();
+    GDALRasterBand* GetMaskBandMainBand() { return m_poMaskBandMainBand; }
     int             IsSameExceptBandNumber( VRTSimpleSource* poOtherSource );
     CPLErr          DatasetRasterIO(
                                GDALDataType eBandDataType,
@@ -899,6 +962,8 @@ public:
 
 class VRTAveragedSource : public VRTSimpleSource
 {
+    CPL_DISALLOW_COPY_ASSIGN(VRTAveragedSource)
+
 public:
                     VRTAveragedSource();
     virtual CPLErr  RasterIO( GDALDataType eBandDataType,
@@ -942,6 +1007,9 @@ typedef enum
 
 class CPL_DLL VRTComplexSource : public VRTSimpleSource
 {
+    CPL_DISALLOW_COPY_ASSIGN(VRTComplexSource)
+    bool           AreValuesUnchanged() const;
+
 protected:
     VRTComplexSourceScaling m_eScalingType;
     double         m_dfScaleOff;  // For linear scaling.
@@ -997,7 +1065,8 @@ public:
                                   void *pProgressData ) override;
 
     virtual CPLXMLNode *SerializeToXML( const char *pszVRTPath ) override;
-    virtual CPLErr XMLInit( CPLXMLNode *, const char *, void* ) override;
+    virtual CPLErr XMLInit( CPLXMLNode *, const char *, void*,
+                            std::map<CPLString, GDALDataset*>& ) override;
     virtual const char* GetType() override { return "ComplexSource"; }
 
     double  LookupValue( double dfInput );
@@ -1024,6 +1093,8 @@ class VRTFilteredSource : public VRTComplexSource
 private:
     int          IsTypeSupported( GDALDataType eTestType ) const;
 
+    CPL_DISALLOW_COPY_ASSIGN(VRTFilteredSource)
+
 protected:
     int          m_nSupportedTypesCount;
     GDALDataType m_aeSupportedTypes[20];
@@ -1054,6 +1125,8 @@ public:
 
 class VRTKernelFilteredSource : public VRTFilteredSource
 {
+    CPL_DISALLOW_COPY_ASSIGN(VRTKernelFilteredSource)
+
 protected:
     int     m_nKernelSize;
 
@@ -1067,7 +1140,8 @@ public:
             VRTKernelFilteredSource();
     virtual ~VRTKernelFilteredSource();
 
-    virtual CPLErr  XMLInit( CPLXMLNode *psTree, const char *, void* ) override;
+    virtual CPLErr  XMLInit( CPLXMLNode *psTree, const char *, void*,
+                             std::map<CPLString, GDALDataset*>& ) override;
     virtual CPLXMLNode *SerializeToXML( const char *pszVRTPath ) override;
 
     virtual CPLErr  FilterData( int nXSize, int nYSize, GDALDataType eType,
@@ -1083,11 +1157,14 @@ public:
 
 class VRTAverageFilteredSource : public VRTKernelFilteredSource
 {
+    CPL_DISALLOW_COPY_ASSIGN(VRTAverageFilteredSource)
+
 public:
             explicit VRTAverageFilteredSource( int nKernelSize );
     virtual ~VRTAverageFilteredSource();
 
-    virtual CPLErr  XMLInit( CPLXMLNode *psTree, const char *, void* ) override;
+    virtual CPLErr  XMLInit( CPLXMLNode *psTree, const char *, void*,
+                             std::map<CPLString, GDALDataset*>& ) override;
     virtual CPLXMLNode *SerializeToXML( const char *pszVRTPath ) override;
 };
 
@@ -1096,11 +1173,14 @@ public:
 /************************************************************************/
 class VRTFuncSource : public VRTSource
 {
+    CPL_DISALLOW_COPY_ASSIGN(VRTFuncSource)
+
 public:
             VRTFuncSource();
     virtual ~VRTFuncSource();
 
-    virtual CPLErr  XMLInit( CPLXMLNode *, const char *, void* ) override { return CE_Failure; }
+    virtual CPLErr  XMLInit( CPLXMLNode *, const char *, void*,
+                             std::map<CPLString, GDALDataset*>& ) override { return CE_Failure; }
     virtual CPLXMLNode *SerializeToXML( const char *pszVRTPath ) override;
 
     virtual CPLErr  RasterIO( GDALDataType eBandDataType,
